from busy.command.command import CollectionCommand
from busy.util.checklist import Checklist


class PrintCommand(CollectionCommand):
    """Generate a Checklist PDF"""

    default_criteria = ["1-"]
    name = "print"

    @CollectionCommand.wrap
    def execute(self):
        checklist = Checklist()
        collection = self.storage.get_collection(
            self.queue, self.collection_state)
        indices = collection.select(*self.criteria)
        if indices:
            items = [collection[i].base for i in indices]
        else:
            self.status = f"Queue '{self.queue}' has " + \
                f"no {self.collection_state} items that meet the criteria"
        queue = self.queue.capitalize()
        state = self.collection_state.capitalize()
        criteria = (": "+",".join(self.criteria)) \
            if (self.criteria != self.default_criteria) else ""
        title = f"{queue} ({state}{criteria})"
        checklist.generate(title, items)



# A Command class handles argparse stuff (ArgumentParser and Namespace) on
# behalf of a method somewhere in the model (i.e. on Queue) and returns text
# that can be used in a UI. A Command class rarely performs an actual
# operation.
#
# Command classes must have the following class-level attributes:
#
# ui - Name of the UI associated with this Command. Default is 'shell', set in
# base class. Inside the command, self._ui refers to the UI object itself.
#
# name - Name of the command. Used in the original Shell UI and also as the
# name of the methods on the queue.
#
# Commands must implement the following methods:
#
# @classmethod set_parser(parser) - Receives an ArgumentParser (actually a
# subparser, but that's irrelevant here), adds arguments to it as appropriate
# for that command, and keeps a reference to it in self.parser.
#
# execute(namespace) - Perform the command. It might use the namespace from the
# Handler, or it might accept a new Namespace, in the event of interactive UIs.
# The execute method is expected to set a property "status" with a one-line
# string description of what happened.
#
# A Command might implement the following methods:
#
# clean_args() - Examine the self._namespace and use the UI to request any args
# that are omitted. Tends to be command-specific.
#
# They might have the following class attributes:
#
# key - For the Curses UI. What key the user should press.
#
# TODO: Move the base class to __init__.py

from argparse import Namespace
from dataclasses import dataclass

from busy.model.collection import Collection
from busy.storage import Storage
from busy.ui.ui import UI
from wizlib.class_family import ClassFamily
from wizlib.super_wrapper import SuperWrapper


@dataclass
class Command(ClassFamily, SuperWrapper):
    """
    Base class for all commands

    Parameters:

    storage - A Busy Storage object (required)

    ui - A Busy UI object
    """

    storage: Storage
    ui: UI = None
    status = ''
    uitype = 'shell'

    @classmethod
    def set_parser(self, parser):
        self.parser = parser

    @classmethod
    def _add_confirmation_arg(self, parser):
        """Add argparse argument for confirmation"""
        parser.add_argument('--yes', '-y', action='store_true')

    def execute(self, method, *args, **kwargs):
        """Actually perform the command, using SuperWrapper"""
        self.clean_args()
        result = method(self, *args, **kwargs)
        return result

    # The clean_args method serves 3 purposes.
    #
    # (1) If any arguments are missing from the command, ask the user to
    # provide them through the UI. For example, when adding a new item to a
    # queue, if a description is not yet provided, ask the user to provide one.
    #
    # (2) This is where defaulting happens. Defaults should generally not be
    # applied in the model, but rather here in the command layer, closer to the
    # user.
    #
    # (3) Call for confirmation from the user (using is_confirmed) if
    # appropriate.
    #
    # The default is to do nothing, as this is essentially an abstract method
    # ripe for override.

    def clean_args(self):
        pass

    def provided(self, argument):
        """Was an argument provided?"""
        value = None
        if hasattr(self, argument):
            value = getattr(self, argument)
        return bool(value)

    def confirm(self, verb, other_action=None):
        """Ensure that a command is confirmed by the user"""
        if self.provided('yes'):
            return self.yes
        else:
            chooser = self.ui.get_chooser(intro=f"{verb}?", default="ok")
            chooser.add_choice(keys=['y', '\n', ''], words=['ok'], action=True)
            chooser.add_choice(keys=['c'], words=['cancel'], action=False)
            if other_action:
                chooser.add_choice(keys=['o'], words=[
                                   'other'], action=other_action)
            choice = self.ui.get_option(chooser)
            if type(choice) == bool:
                self.yes = choice
            return choice


@dataclass
class QueueCommand(Command):
    """Base for commands that work on the default collection of one queue"""

    queue: str = 'tasks'
    collection_state: str = 'todo'
    criteria: list = None
    default_criteria = [1]

    @property
    def collection(self):
        """Return the collection object being queried"""
        if not hasattr(self, '_collection'):
            self._collection = self.storage.get_collection(
                self.queue, self.collection_state)
        return self._collection

    @property
    def selection(self):
        """Indices of objects within the query collection that match the
        criteria"""
        if not hasattr(self, '_selection'):
            self._selection = self.collection.select(*self.criteria)
        return self._selection

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('queue', default='tasks', nargs='?')
        parser.add_argument('--criteria', '-c', action='store', nargs="*")

    def clean_args(self):
        """Apply default criteria"""
        super().clean_args()
        if not self.provided('criteria'):
            self.criteria = self.default_criteria

    @Command.wrap
    def execute(self, method, *args, **kwargs):
        """Execute the command then save the collection(s)"""
        result = method(self, *args, **kwargs)
        self.storage.save()
        return result

    def count(self, items=None):
        """Return a friendly string count of some items"""
        if items is None:
            items = self.selection
        if len(items) == 1:
            return "1 item"
        if len(items) > 1:
            return str(len(items)) + " items"
        return "nothing"


@dataclass
class CollectionCommand(QueueCommand):
    """Base for commands that work on a user-specified collection"""

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        states = Collection.family_attrs('state')
        parser.add_argument(
            '--state', '-s', action='store', default='todo',
            dest='collection_state', choices=states)

    def output_items(self, func, with_index=False):
        """Return some attribute of all the items in the collection"""
        collection = self.storage.get_collection(
            self.queue, self.collection_state)
        indices = collection.select(*self.criteria)
        if indices:
            if with_index:
                return '\n'.join([func(collection[i], i) for i in indices])
            else:
                return '\n'.join([func(collection[i]) for i in indices])
        else:
            self.status = f"Queue '{self.queue}' has " + \
                f"no {self.collection_state} items"



from dataclasses import dataclass

from busy.command.command import QueueCommand
from busy.util import date_util


def is_today_or_earlier(plan):
    return plan.plan_date <= date_util.today()


@dataclass
class ActivateCommand(QueueCommand):

    timing: str = ""
    yes: bool = None
    collection_state: str = 'plan'
    name = 'activate'
    key = 'c'
    default_criteria = [is_today_or_earlier]

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        # parser.add_argument('--timing', '-t', default='today')
        self._add_confirmation_arg(parser)

    def clean_args(self):
        # def update_timing():
        #     self.timing = self.ui.get_string("Timing", "today")
        super().clean_args()
        # For now, handling criteria as normal
        if self.selection:
            if not self.provided('yes'):
                items = self.collection.items(self.selection)
                self.ui.output('\n'.join([str(i) for i in items]))
                intro = f"Activate {self.count()}"
                # self.confirm(intro, update_timing)
                self.confirm(intro)

    @QueueCommand.wrap
    def execute(self):
        if not self.selection:
            self.status = "Activated nothing"
        elif self.yes is False:
            self.status = "Activate command canceled"
        else:
            todos = self.storage.get_collection(self.queue)
            activated = self.collection.delete(self.selection)
            for item in activated:
                item.state = 'todo'
            todos[0:0] = activated
            self.status = f"Activated {self.count()}"



from ..command.command import Command


class CursesCommand(Command):

    name = 'curses'
    uitype = 'curses'

    def execute(self):
        """Start the Curses UI"""
        self.ui.start()
        print(f'Session complete: {self.ui.name}')



from dataclasses import dataclass

from busy.command.command import QueueCommand


@dataclass
class SwitchCommand(QueueCommand):
    """
    Switch to a different queue. Note this only works in interactive UIs; the
    shell UI has no state so it's essentially a no-op.
    """

    name = "switch"
    key = "w"
    yes: bool = None

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_confirmation_arg(parser)

    def clean_args(self):
        self.newqueuename = self.ui.get_string("Switch to queue", "tasks")
        if not self.provided('yes'):
            if not self.storage.queue_exists(self.newqueuename):
                self.confirm(f"Switch to empty queue {self.newqueuename}")
            else:
                self.yes = True

    @QueueCommand.wrap
    def execute(self):
        if self.yes:
            self.queue = self.newqueuename
            self.status = f"Switched to queue {self.newqueuename}"
        else:
            self.status = "Switch operation canceled"



from dataclasses import dataclass

from busy.command.command import CollectionCommand


@dataclass
class DeleteCommand(CollectionCommand):

    yes: bool = None
    name = 'delete'
    key = 'd'
    default_criteria = [1]

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--yes', action='store_true', default=None)

    def clean_args(self):
        super().clean_args()
        if self.selection:
            items = self.collection.items(self.selection)
            self.ui.output('\n'.join([str(i) for i in items]))
            self.confirm(f"Delete {self.count()}")

    # Assume the indices have been already set, before confirmation.

    @CollectionCommand.wrap
    def execute(self):
        if self.yes:
            deleted = self.collection.delete(self.selection)
            self.status = f"Deleted {self.count(deleted)}"
        else:
            self.status = "Delete command canceled"



# The Handler is the core control class for a single execution of the busy
# program. It's called when busy is run, and there are some functions to get
# things configured and started. Then one Handler is instantiated. This is the
# only place that imports the UI, Command, and Root classes. So the handler
# instance is passed around, and other classes can access what they need from
# there. Keeping the imports in one place allows the class families to be more
# modular.
#
# A Handler instance provides the following properties. For the most part,
# these
# properties are how commands access other class families. Remember that the
# Namespace from the Handler is the original Namespace, so use the one passed
# in to the execute() method of the Command for actual usage in an interactive
# UI.


import sys
from argparse import ArgumentParser
from pathlib import Path
from sys import argv

from busy.command.command import Command
from busy.storage.file_storage import FileStorage
from busy.ui.ui import UI, UserCancelError
from busy.util.python_version import confirm_python_version

PYTHON_VERSION = (3, 6, 5)


def main(*args):
    """Start everything from the command line

    args - optional entry point for testing
    """
    args = args if args else argv[1:]
    confirm_python_version(PYTHON_VERSION)
    parser = get_parser()
    namespace = parser.parse_args(args)
    try:
        handler = Handler(namespace)
        handler.execute()
    except (RuntimeError, ValueError) as error:
        print(f"Error: {str(error)}")


def get_parser():
    """Return the CLI parser"""
    parser = ArgumentParser(prog='busy')
    parser.add_argument('--root', action='store',
                        help='file system path to busy data',
                        default=None)
    subparsers = parser.add_subparsers(dest='command')
    for command in Command.family_members('name'):
        key = command.get_member_attr('key')
        aliases = [key] if key else []
        subparser = subparsers.add_parser(command.name, aliases=aliases)
        command.set_parser(subparser)
    return parser


class Handler():

    def __init__(self, namespace):
        """
        namespace - ArgParse Namespace from initial CLI entry
        """
        nsvars = vars(namespace)
        self.storage = FileStorage(nsvars.pop('root'))
        command = nsvars.pop('command')
        command = command if command else 'curses'
        command_class = Command.family_member('name', command)
        if not command_class:
            raise RuntimeError(f"Unknown command {command}")
        self.ui = UI.family_member('name', command_class.uitype)(self)
        self.command = command_class(
            storage=self.storage, ui=self.ui, **nsvars)

    def execute(self):
        """Handle the command and output its final status"""
        result = self.command.execute()
        if result:
            print(result)
        if self.command.status:
            print(self.command.status)

    # Some useful methods so other components don't need to import from
    # different class families

    def get_commands(self, *attributes):
        return Command.family_members(*attributes)

    # A method for commands within interactive UIs. Use our own storage but the
    # UI is passed in. Returns None if attribute value doesn't resolve.

    def get_command(self, attribute, value, **kwargs):
        member = Command.family_member(attribute, value)
        if member:
            args = {'storage': self.storage} | kwargs
            return member(**kwargs)



# Prototype of how to generate a PDF

from subprocess import run
from pathlib import Path
from dataclasses import dataclass
from tempfile import TemporaryDirectory
from datetime import date
import re
from time import sleep

from reportlab.lib.pagesizes import letter as LETTER
from reportlab.lib.units import inch as INCH
from reportlab.pdfgen.canvas import Canvas
from reportlab.pdfbase import pdfmetrics


@dataclass
class Checklist:

    filenamestem: str = None
    pagesize: tuple = LETTER
    margin: float = 1.0 * INCH
    fontname_items: str = "Helvetica"
    fontname_heading: str = "Helvetica-Bold"
    fontsize_items: float = 12.0
    fontsize_heading: float = 24.0
    lineheight_items: float = 14.0
    paragap: float = 8.0
    checkbox_size: float = 14.0

    @property
    def topy(self):
        return self.pagesize[1] - self.margin

    @property
    def indent(self):
        return self.checkbox_size * 1.5

    def wrap(self, text: str) -> list:
        textwidth = self.pagesize[0] - 2.0 * self.margin - self.indent
        lines = []
        words = text.split()
        line = ""
        while words:
            word = words.pop(0)
            newwidth = pdfmetrics.stringWidth(line + " " + word,
                                              self.fontname_items,
                                              self.fontsize_items)
            if newwidth <= textwidth:
                line += " " + word
            else:
                lines.append(line)
                line = word
        lines.append(line)
        return lines

    def print_checkbox(self, canvas: Canvas, x, y):
        """y is the baseline"""
        ascent = pdfmetrics.getAscent(self.fontname_items, self.fontsize_items)
        y = y - (self.lineheight_items - ascent) / 2.0
        canvas.rect(x, y, self.checkbox_size, self.checkbox_size)

    def print(self, canvas: Canvas, text: str, y) -> int:
        """Add an item to the document and return the new y-coordinate"""
        lines = self.wrap(text)
        if y - len(lines) * self.lineheight_items < self.margin:
            canvas.showPage()
            heading = f"Page {canvas.getPageNumber()}"
            canvas.setFont(self.fontname_heading, self.fontsize_items)
            canvas.drawString(self.margin, self.topy, heading)
            y = self.topy - self.lineheight_items - self.paragap
        self.print_checkbox(canvas, self.margin, y)
        canvas.setFont(self.fontname_items, self.fontsize_items)
        while lines:
            line = lines.pop(0)
            canvas.drawString(self.margin + self.indent, y, line)
            y = y - self.lineheight_items
        return y - self.paragap

    def print_pdf(filepath):
        fullpath = Path(filepath).absolute()
        applescript_cmd = f'tell application "Preview" to ' + \
            f'print POSIX file "{fullpath}"'
        run(['osascript', '-e', applescript_cmd])

    def generate(self, heading: str, items: list):
        datestring = date.today().strftime('%Y%m%d')
        if not (stem := self.filenamestem):
            stem1 = re.sub(r'\W+', '-', heading.lower())
            stem = re.sub(r'\-$', '', stem1)
        with TemporaryDirectory() as dir:
            filepath = Path(dir) / f"{datestring}-{stem}.pdf"
            canvas = Canvas(str(filepath), pagesize=self.pagesize)
            canvas.setFont(self.fontname_heading, self.fontsize_heading)
            canvas.drawString(self.margin, self.topy, heading)
            y = self.topy - self.fontsize_heading - self.paragap
            for item in items:
                y = self.print(canvas, item.capitalize(), y)
            canvas.save()
            run(['open', filepath.absolute()])
            sleep(1)  # A hack to give Preview time to open the temp file


if __name__ == '__main__':

    items = [
        "welcome mat",
        "wall patch kit",
        "ReportLab can handle word wrap and text flow in generated PDF docs",
        "brooms (2)",
        "bins",
        """
        When using ReportLab, you have fine-grained control over the layout and
        formatting of text. You can specify the width of text areas and enable
        automatic word wrapping within those areas. This ensures that the text
        flows naturally within the defined boundaries, wrapping to the next
        line when necessary.
        """,
        "Bananas"
    ] * 8

    title = "Shopping"

    p = Checklist()
    p.generate(title, items)



import re
from dataclasses import KW_ONLY, dataclass
from datetime import date

from busy.util.date_util import absolute_date, relative_date


class ItemStateError(Exception):
    pass


@dataclass
class Item:

    description: str
    _: KW_ONLY
    state: str = 'todo'
    done_date: date = None
    plan_date: date = None

    def __str__(self):
        """Represent the item as its description"""
        return self.description

    def restricted(*allowed_states):
        """Restrict a method to a specific set of states"""
        def wrapper(method):
            def replacement(self, *args, **kwargs):
                if self.state in allowed_states:
                    return method(self, *args, **kwargs)
                else:
                    raise ItemStateError
            return replacement
        return wrapper

    FOLLOW_SPLIT = re.compile(r'\s*\-*\>\s*')
    REPEAT = re.compile(r'^\s*repeat(?:\s+[io]n)?\s+(.+)\s*$', re.I)

    def __setattr__(self, name, value):
        if self.__annotations__[name] == date:
            value = absolute_date(value)
        super().__setattr__(name, value)

    @property
    def first(self):
        return self.FOLLOW_SPLIT.split(self.description, maxsplit=1)[0]

    @property
    def _words(self):
        return self.first.split()

    @property
    def tags(self):
        wins = [w for w in self._words if w.startswith("#")]
        return {w[1:].lower() for w in wins}

    @property
    def resource(self):
        wins = [w for w in self._words if w.startswith("@")]
        return wins.pop()[1:] if wins else ""

    @property
    def base(self):
        wins = [w for w in self._words if w[0] not in '#@']
        return " ".join(wins)

    @property
    def next(self):
        """Second and successive segments, all but current"""
        split = self.FOLLOW_SPLIT.split(self.description, maxsplit=1)
        if len(split) > 1:
            return split[1]
        else:
            return ""

    @property
    def current(self):
        """The first segment, all but next"""
        split = self.FOLLOW_SPLIT.split(self.description, maxsplit=1)
        if split:
            return split[0]
        else:
            return ""

    @property
    def repeat(self):
        followon = self.next
        match = self.REPEAT.match(followon)
        if match:
            return match.group(1)
        else:
            return ""

    @property
    def repeat_date(self):
        repeat = self.repeat
        if repeat:
            return relative_date(repeat)

    @restricted('todo')
    def done(self, done_date: date):
        self.state = 'done'
        self.done_date = done_date
        if self.repeat:
            follow = Item(self.description, state='plan')
            follow.plan_date = relative_date(self.repeat)
            return follow
        if self.next:
            follow = Item(self.next)
            self.description = self.first
            return follow

    @restricted('done')
    def undone(self):
        self.state = 'todo'

    @restricted('todo')
    def skip(self):
        self.state = 'skip'

    @restricted('skip')
    def unskip(self):
        self.state = 'todo'

    @restricted('todo')
    def plan(self, plan_date: date):
        self.state = 'plan'
        self.plan_date = plan_date

    @restricted('plan')
    def unplan(self):
        self.state = 'todo'



from dataclasses import dataclass

from busy.command.command import CollectionCommand


@dataclass
class EditorCommandBase(CollectionCommand):

    @CollectionCommand.wrap
    def execute(self):
        if not self.selection:
            self.status = "Edited nothing"
        else:
            edited = self.ui.edit_items(self.collection, self.selection)
            self.status = f"Edited {self.count(edited)}"


# Special case versions of EditCommand to edit just the top item or the whole
# queue. In the shell you'd just type "edit" or "edit 1-" but this is
# convenient in curses.

class EditCommand(EditorCommandBase):

    name = "edit"
    key = "e"


class EditAllCommand(EditorCommandBase):

    name = 'manage'
    key = 'm'
    default_criteria = ["1-"]



from busy.command.command import CollectionCommand


class ResourceCommand(CollectionCommand):
    """Get the resource (URL) for items in a collection"""

    name = "resource"

    @CollectionCommand.wrap
    def execute(self):
        return self.output_items(lambda i: i.resource)



# Abstract classes for UIs. The idea here is to allow many different user
# interfaces (shell, curses, Slack, etc) to drive the same data model without
# the data model needing specific UI knowledge.
#
# The UI keeps a reference to the Handler. The Handler will bring a Command
# with it, and this is the main root command. In the case of a Shell-type UI,
# that is the command to be executed. In the case of an interactive UI, it's
# really just a placeholder; the UI will instantiate other commands as it
# proceeds. But the Handler won't know about those.
#
# Commands might call back to the UI for confirmations or arguments that are
# previously omitted, using the get_ methods.

#
# UI end classes must implement the following interface:
#
# __init__(handler): Takes the handler and hangs on to it. Implemented in the
# main base class.
#
# start(): No arguments. Actually performs the operation of the UI. It might be
# short running (in the case of a shell UI) or long-running (in the case of an
# interactive UI).
#
# output(intro=""): Output some multi-line text explaining an action, usually a
# list of items being acted upon.
#
# get_string(prompt, default=""): For arguments that are omitted, get a string
# from the user. The prompt is just a word (like "Description") telling the
# user what to input.
#
# get_confirmation(verb): For delete-oriented commands to confirm with the user
# before proceeding. Verb is what we're asking the user to confirm. Description
# can be multiple lines, as with get_string. Returns a boolean saying whether
# the action is confirmed.

import os
import re
import shutil
import subprocess
from dataclasses import dataclass, field
from enum import Enum
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile

from busy.model.collection import Collection
from wizlib.class_family import ClassFamily


@dataclass
class Choice:
    """One option of several"""

    keys: list[str] = field(default_factory=list)
    words: list[str] = field(default_factory=list)
    action: object = None

    @property
    def key(self):
        return self.keys[0]

    @property
    def word(self):
        return self.words[0]

    def hit_key(self, key):
        return key in self.keys

    def hit_word(self, word):
        return word in self.words

    def call(self, *args, **kwargs):
        if callable(self.action):
            return self.action(*args, **kwargs)


@dataclass
class Prompt:
    """Tell the user what kind of input to provide"""

    intro: str = ""
    default: str = ""
    choices: list = field(default_factory=list)


@dataclass
class Chooser(Prompt):
    """Hold a set of choices and get the result"""

    def add_choice(self, *args, **kwargs):
        choice = Choice(*args, **kwargs)
        self.choices.append(choice)

    def choice_by_key(self, key):
        chosenlist = [o.action for o in self.choices if o.hit_key(key)]
        return self._choose(chosenlist)

    def choice_by_word(self, word):
        chosenlist = [o.action for o in self.choices if o.hit_word(word)]
        return self._choose(chosenlist)

    def _choose(self, chosenlist):
        choice = next(iter(chosenlist), None)
        if callable(choice):
            return choice()
        else:
            return choice


@dataclass
class UI(ClassFamily):

    handler: object

    def get_chooser(self, *args, **kwargs):
        """Return a chooser that can be set up with options"""
        return Chooser(*args, **kwargs)

    # TODO: Make start an abstract method
    def start(self):
        pass

    # TODO: Make get_option an abstract method
    def get_option(self, chooser: Chooser):
        """Get an option from a list from the user"""
        pass

    # TODO: Make get_string an abstract method
    def get_string(self, intro, default=""):
        """Get a string value from the user"""
        pass

    def edit_items(self, collection: Collection, indices):
        pass

# Terminal UIs include Shell and Curses, since both will rely on a
# terminal-based editor for the e.g. Manage command.


class TerminalUI(UI):

    # A convenience method to get a full textual prompt for a string input

    def full_prompt(self, prompt, default=None):
        if default:
            return f'{prompt} [{default}]: '
        else:
            return f'{prompt}: '

    @staticmethod
    def edit_text(text):
        commands = [['sensible-editor'], ['open', '-W']]
        with NamedTemporaryFile(mode="w+") as tempfile:
            tempfile.write(text)
            tempfile.seek(0)
            command = [os.environ.get('EDITOR')]
            if not command[0] or not shutil.which(command[0]):
                iterator = (c for c in commands if shutil.which(c[0]))
                command = next(filter(None, iterator), None)
                if not command:
                    raise RuntimeError(
                        "A text editor at the $EDITOR " +
                        "environment variable is required")
            subprocess.run(command + [tempfile.name])
            tempfile.seek(0)
            return tempfile.read()

    def edit_items(self, collection: Collection, indices):
        with StringIO() as oldio:
            collection.write_items(oldio, indices)
            oldio.seek(0)
            oldtext = oldio.read()
        newtext = self.edit_text(oldtext)
        with StringIO() as newio:
            newio.write(newtext)
            newio.seek(0)
            newitems = collection.read_items(newio, indices)
        return newitems


class UserCancelError(Exception):
    pass



# A terminal-based UI uses Curses, which offers one-key access to most common
# operations.
#
# The terminal has 3 regions, implemented as curses window objects. The regions
# are organized top-to-bottom with a 1-row gap between them. There is also a
# 1-row/1-column gap around the outside of the main terminal window containing
# the regions.
#
# Status region (_statuswin) - 6 lines at the top. Current queue name, current
# task, some blank space, status of the most recent operation, and input prompt
# for the main command.
#
# List region (_listwin) - Takes up the space between the other two regions.
# Shows mostly lists of tasks for confirmation and/or management.
#
# Edit region (_editwin) - 2 lines at the bottom. Used for editing of
# individual task descriptions, such as when adding a new task to the queue.


import curses
from argparse import Namespace
from string import capwords
from time import sleep

from busy.ui.ui import Chooser, Prompt, TerminalUI, UserCancelError


class CursesError(Exception):
    pass


class CursesUI(TerminalUI):  # pragma: nocover

    name = "curses"

    def start(self):
        self._mode = "WORK"
        chooser = Chooser()
        commands = self.handler.get_commands('key')
        scommands = sorted(commands, key=lambda c: c.name)
        for command in scommands:
            chooser.add_choice(
                keys=[command.key],
                words=[command.name],
                action=command
            )
        chooser.add_choice(keys=['q'], words=['quit'])
        self._command_prompt = chooser

        curses.wrapper(self.term_loop)

    def output(self, intro=""):
        self._descwin.clear()
        self._descwin.addstr(intro)
        self._descwin.refresh()

    def write_prompt(self, prompt):
        """
        Output the prompt with underlines to the window.
        """
        window = self._statuswin
        if prompt.intro:
            window.addstr(prompt.intro + " ", curses.color_pair(3))
        if prompt.default:
            window.addstr(f"[{prompt.default}] ", curses.color_pair(3))
        for choice in prompt.choices:  # TODO: Deal with raw prompt case
            if choice.word == prompt.default:
                continue
            pre, it, post = choice.word.partition(choice.key)
            window.addstr(pre, curses.color_pair(3))
            window.addstr(it, curses.A_UNDERLINE + curses.color_pair(3))
            window.addstr(post + " ", curses.color_pair(3))
        window.addstr(": ", curses.color_pair(3))

    # TODO: Use a better editing component
    #
    # NOTE: get_string assumes that everything has been cleared and we are in
    # the right place.

    def get_string(self, intro, default=""):
        prompt = Prompt(intro=intro, default=default)
        self._update_status(prompt)
        curses.echo()
        try:
            string = self._statuswin.getstr()
        except KeyboardInterrupt:
            raise UserCancelError
        value = string.decode() or default
        curses.noecho()
        return value

    def get_option(self, chooser):
        """Get a 1-keystroke choice from the user"""
        self._update_status(chooser)
        key = self._get_key()
        return chooser.choice_by_key(key)

    def term_loop(self, fullwin):
        self.queue = "tasks"
        curses.init_color(curses.COLOR_BLACK, 200, 200, 200)
        curses.init_pair(1, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_GREEN, curses.COLOR_BLACK)
        h, w = fullwin.getmaxyx()
        self._statuswin = curses.newwin(5, w-2, 1, 1)
        self._descwin = curses.newwin(h-8, w-2, 7, 1)
        self._editwin = curses.newwin(1, w-2, h-2, 1)
        self._status = "Welcome to Busy!"
        list = ""
        cursor = 0
        while True:
            self.output()
            self._update_status(self._command_prompt)
            try:
                key = self._get_key()
            except UserCancelError:
                break
            except CursesError:
                sleep(1)
                continue
            if key == "q":
                break
            command = self.handler.get_command(
                'key', key, ui=self, storage=self.handler.storage,
                queue=self.queue)
            if not command:
                self._status = f"Invalid command {key}"
                continue
            command_name = capwords(command.name)
            self._status = f"{command_name} in progress"
            self._update_status()
            try:
                result = command.execute()
                self._status = command.status or ""
                if hasattr(command, 'queue'):
                    self.queue = command.queue
            except UserCancelError:
                self._status = f"{command_name} command canceled"

    def _update_status(self, prompt=None):
        collection = self.handler.storage.get_collection(self.queue)
        self._statuswin.clear()
        self._statuswin.move(0, 0)
        self._statuswin.addstr("Queue: ", curses.color_pair(1))
        self._statuswin.addstr(capwords(self.queue), curses.A_BOLD)
        self._statuswin.move(1, 0)
        self._statuswin.addstr("Todo:   ", curses.color_pair(1))
        self._statuswin.addstr(
            str(collection[0] if collection else ''), curses.A_BOLD)
        self._statuswin.move(3, 0)
        self._statuswin.addstr(self._status, curses.color_pair(2))
        self._statuswin.move(4, 0)
        if prompt:
            self.write_prompt(prompt)
        cursor = self._statuswin.getyx()
        self._statuswin.refresh()
        self._statuswin.move(*cursor)

    # Convenience method to get one keystroke from the user.
    #
    def _get_key(self):
        try:
            key = self._statuswin.getkey()
        except KeyboardInterrupt:
            raise UserCancelError
        except curses.error:
            raise CursesError
        return key

    # def listmode(self):
    #     length = len(list.splitlines())
    #     if length:
    #         listpad = curses.newpad(length, curses.COLS)
    #         listpad.addstr(list)
    #         self.listmode = True
    #         listpad.move(cursor, 0)
    #         listpad.addstr(">")
    #         listpad.refresh(0, 0, 5, 0, curses.LINES - 5, curses.COLS)
    #     if self.listmode:
    #         if key == "KEY_UP" and cursor > 0:
    #             cursor = cursor - 1
    #         if key == "KEY_DOWN" and cursor < length:
    #             cursor = cursor + 1



from dataclasses import dataclass

from busy.command.command import QueueCommand
from busy.model.item import Item


@dataclass
class AddCommand(QueueCommand):

    description: str = ""
    name = 'add'
    key = 'a'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--description', '-d', nargs='?')

    def clean_args(self):
        super().clean_args()
        if not self.provided('description'):
            self.description = self.ui.get_string("Description")

    @QueueCommand.wrap
    def execute(self):
        if self.description:
            item = Item(self.description)
            self.collection.insert(0, item)
            self.status = "Added: " + self.description
        else:
            self.status = "Nothing added"
